/*
 * Copyright (C) 2004 by kabalNET.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * All copyright notices regarding KNCwebserver MUST remain intact.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * @author Brett Haas
 */
package com.kabalnet.kncwebserver.core;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import java.text.DecimalFormat;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;

import java.util.Calendar;

public class Logger {
    private static final String DATE_FORMAT = "yyyy/MM/dd k:m:s.S Z";
    private static final Object lock = new Object();
    private Calendar cal = Calendar.getInstance();
    private NumberFormat numberFormatter = new DecimalFormat("00");
    private PrintWriter pw;
    private SimpleDateFormat dateFormatter;
    private String dateFormat;
    private String dir;
    private String fileName;
    private String logDate;
    private String messageFormat;

    /**
     * Init Object
     *
     * @param dir Directory to place log files
     * @param fileName Name of the log file
     * @param messageFormat java.text.MessageFormat pattern.
     */
    public Logger(String dir, String fileName, String messageFormat) {
        this(dir, fileName, messageFormat, DATE_FORMAT);
    }

    /**
     * Init Object
     *
     * @param dir Directory to place log files
     * @param fileName Name of the log file
     * @param messageFormat java.text.MessageFormat pattern
     * @param dateFormat java.text.SimpleDateFormat pattern
     */
    public Logger(String dir, String fileName, String messageFormat, String dateFormat) {
        this.dir = dir;
        verifySetupParam(dir, "directory");

        this.fileName = fileName;
        verifySetupParam(fileName, "file name");

        this.messageFormat = messageFormat;
        verifySetupParam(messageFormat, "message format");

        this.dateFormat = dateFormat;
        verifySetupParam(dateFormat, "date format");

        dateFormatter = new SimpleDateFormat(dateFormat);

        init();
    }

    /**
     * Logs out items to text file
     *
     * @param itemsToLog String[]
     */
    public void log(String[] itemsToLog) {
        if (itemsToLog == null) {
            throw new IllegalArgumentException("itemsToLog can not be null");
        }

        if (pw != null) {
            synchronized (lock) {
                if (!getDate().equals(logDate)) {
                    System.out.println("Creating new log due to date change");
                    init();
                }
            }

            Object[] objects = new Object[itemsToLog.length + 1];
            objects[0] = dateFormatter.format(cal.getTime());

            for (int i = 1; i < objects.length; i++) {
                objects[i] = itemsToLog[i - 1];
            }

            pw.println(MessageFormat.format(messageFormat, objects));
        }
    }

    private String getDate() {
        cal.setTimeInMillis(System.currentTimeMillis());

        int year = cal.get(Calendar.YEAR);
        int month = cal.get(Calendar.MONTH) + 1;
        int day = cal.get(Calendar.DATE);
        return year + "-" + numberFormatter.format(month) + "-" + numberFormatter.format(day);
    }

    private void init() {
        logDate = getDate();

        try {
            pw = new PrintWriter(new BufferedWriter(
                        new FileWriter(
                            new File(dir + File.separatorChar + fileName + logDate + ".txt"),
                            true)), true);
        } catch (IOException ex) {
            System.out.println("Could not enable logging. Error: " + ex.getMessage());
        }
    }

    private void verifySetupParam(String value, String desc) {
        if (value == null || value.trim().length() == 0) {
            throw new IllegalArgumentException("Arg (" + desc + ") can not be null or empty");
        }
    }
}
