/*
 * Copyright (C) 2004 by kabalNET.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * All copyright notices regarding KNCwebserver MUST remain intact.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * @author Brett Haas
 */
package com.kabalnet.kncwebserver.core;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;

import java.net.ServerSocket;
import java.net.Socket;
import java.net.URLDecoder;

import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class WebServer extends Thread {
    public static final String VERSION = "KNCwebserver-1.2.3";
    private static final long MAX_THREAD_WAIT_TIMEOUT = 5000;
    private static final String CLIENT_REQUEST_PATTERN = "^GET (.+) HTTP/.+$";
    private static final String CRLF = "\r\n";
    private static final String DEFAULT_MIME_TYPE = "application/octet-stream";
    private static final String LOG_FLAG = "-log";
    private static final String CONFIG_DIR = System.getProperty("config.dir");
    private static final String HTTPD_FILE = "httpd.properties";
    private static final String MIME_TYPES_FILE = "mime-types.properties";
    private static final String LOG_NAME_PREFIX = "access_log.";
    private static final String DIR_LIST_FILE = "dir_list.html";
    private static final String ERROR_FILE = "error.html";
    private static final String END_LOOP_DIRECTIVE = "${loop:end}";
    private static final String DIRECTORY_DIRECTIVE = "\\$\\{directory}";
    private static final String ITEM_LINK_DIRECTIVE = "\\$\\{item_link}";
    private static final String ITEM_NAME_DIRECTIVE = "\\$\\{item_name}";
    private static final String REQUESTED_URL_DIRECTIVE = "\\$\\{requested_url}";
    private static final String START_LOOP_DIRECTIVE = "${loop:start}";
    private static final String STATUS_CODE_DIRECTIVE = "\\$\\{status_code}";
    private static final String STATUS_CODE_MESSAGE_DIRECTIVE = "\\$\\{status_code_message}";
    private static final String BAD_STATUS_CODE = "400";
    private static final String FORBIDDEN_STATUS_CODE = "403";
    private static final String NOT_FOUND_STATUS_CODE = "404";
    private static final String OK_STATUS_CODE = "200";
    private static final Map statusCodes;

    static {
        Map tmp = new HashMap();
        tmp.put(BAD_STATUS_CODE, "Bad Request");
        tmp.put(FORBIDDEN_STATUS_CODE, "Forbidden");
        tmp.put(NOT_FOUND_STATUS_CODE, "Not Found");
        tmp.put(OK_STATUS_CODE, "OK");
        statusCodes = Collections.unmodifiableMap(tmp);
    }

    private Config config;
    private Logger logger;
    private Pattern pattern;
    private String dirListTemplate;
    private String errorTemplate;
    private boolean isLogging;
    private boolean useDirListPage;
    private boolean useErrorPage;

    /**
     * Init Object with logging turned off
     */
    public WebServer() {
        this(false);
    }

    /**
     * Init Object
     *
     * @param isLogging true/false
     */
    public WebServer(boolean isLogging) {
        this.isLogging = isLogging;
        config = new Config();

        loadHttpdFile();
        loadMimeTypesFile();
        loadErrorPage();
        loadDirListPage();

        logger =
            (isLogging)
            ? new Logger(config.getLogsDir(), LOG_NAME_PREFIX, config.getLogMessageFormat()) : null;
        pattern = Pattern.compile(CLIENT_REQUEST_PATTERN);
    }

    public static void main(String[] args) {
        WebServer webserver;

        if (args != null && args.length > 0 && args[0] != null && args[0].equals(LOG_FLAG)) {
            webserver = new WebServer(true);
        } else {
            webserver = new WebServer();
        }

        webserver.start();
    }

    public void run() {
        System.out.println("Starting " + VERSION + " ... Listening on port "
            + config.getPortNumber());

        ServerSocket serverSocket = null;

        try {
            ThreadPool pool = new ThreadPool(config.getPoolSize(), MAX_THREAD_WAIT_TIMEOUT);

            serverSocket = new ServerSocket(config.getPortNumber());

            while (true) {
                Socket socket = serverSocket.accept();
                pool.putTask(new SocketTask(socket));
            }
        } catch (Exception ex) {
            System.out.println("Server died unexpectedly. Error: " + ex.getMessage());
        } finally {
            try {
                if (serverSocket != null) {
                    serverSocket.close();
                }
            } catch (IOException ex) {
                ;
            }
        }
    }

    private Properties loadConfig(String fileName) throws Exception {
        Properties properties = new Properties();
        properties.load(new FileInputStream(
                new File(CONFIG_DIR + File.separatorChar + fileName).getCanonicalPath()));
        return properties;
    }

    private void loadDirListPage() {
        if (config.isDirBrowsingPermitted()) {
            dirListTemplate = loadTemplate(DIR_LIST_FILE);
            useDirListPage = (!dirListTemplate.equals(""));
        }
    }

    private void loadErrorPage() {
        errorTemplate = loadTemplate(ERROR_FILE);
        useErrorPage = (!errorTemplate.equals(""));
    }

    private void loadHttpdFile() {
        try {
            Properties properties = loadConfig(HTTPD_FILE);

            config.setDocumentRoot(properties.getProperty("document.root"));
            config.setDirBrowsingPermitted(new Boolean(properties.getProperty(
                        "dir.browsing.permitted")).booleanValue());
            config.setPoolSize(Integer.parseInt(properties.getProperty("pool.size")));
            config.setPortNumber(Integer.parseInt(properties.getProperty("port.number")));
            config.setLogsDir(properties.getProperty("logs.dir"));
            config.setTemplatesDir(properties.getProperty("templates.dir"));
            config.setLogMessageFormat(properties.getProperty("log.message.format"));
            config.setIndexPage(properties.getProperty("index.page"));
        } catch (Exception ex) {
            System.out.println("Could not load httpd file. Error: " + ex.getMessage());
            System.exit(0);
        }
    }

    private void loadMimeTypesFile() {
        try {
            Properties properties = loadConfig(MIME_TYPES_FILE);
            Enumeration e = properties.propertyNames();
            Map mimeTypes = new HashMap();

            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                String value = properties.getProperty(key);

                mimeTypes.put(key, value);
            }

            config.setMimeTypes(mimeTypes);
        } catch (Exception ex) {
            System.out.println("Could not load mime types file. Error: " + ex.getMessage());
            System.exit(0);
        }
    }

    private String loadTemplate(String template) {
        String contents = "";
        String errorMessage = "null";

        try {
            File file = new File(config.getTemplatesDir() + File.separatorChar + template);

            if (file.exists()) {
                BufferedReader br = new BufferedReader(new FileReader(file));
                StringBuffer lines = new StringBuffer();

                char[] buf = new char[5000];
                int bufLen = buf.length;
                int offset = 0;

                while ((offset = br.read(buf, 0, bufLen)) != -1) {
                    lines.append(buf, 0, offset);
                }

                contents = lines.toString();
            } else {
                errorMessage = "File does not exist";
            }
        } catch (Exception ex) {
            errorMessage = ex.getMessage();
        }

        if (contents.equals("")) {
            System.out.println("Could not load " + template + " template. Error: " + errorMessage);
        }

        return contents;
    }

    private class SocketTask implements Runnable {
        private Socket socket;

        public SocketTask(Socket socket) {
            this.socket = socket;
        }

        public void run() {
            BufferedReader clientRequest = null;
            BufferedInputStream fileInputStream = null;
            DataOutputStream serverResponse = null;

            try {
                if (socket != null && !socket.isClosed()) {
                    String line;
                    clientRequest =
                        new BufferedReader(new InputStreamReader(socket.getInputStream()));

                    if ((line = clientRequest.readLine()) != null) {
                        String contentType = "text/html";
                        String requestFile = "";
                        String unparsedRequestFile = "";
                        String statusCode = BAD_STATUS_CODE;

                        try {
                            Matcher matcher = pattern.matcher(line);
                            matcher.matches();
                            unparsedRequestFile = matcher.group(1);
                            requestFile = unparsedRequestFile;
                            statusCode = OK_STATUS_CODE;
                        } catch (Exception ex) {
                            ;
                        }

                        StringBuffer headers = new StringBuffer();
                        File file = null;

                        headers.append("HTTP/1.1 ");

                        if (statusCode.equals(OK_STATUS_CODE)) {
                            requestFile = parseRequestFile(requestFile);

                            file = new File(config.getDocumentRoot() + requestFile);

                            if (!file.exists()) {
                                statusCode = NOT_FOUND_STATUS_CODE;
                            } else {
                                if (file.isDirectory()) {
                                    File testFile = new File(file.getCanonicalPath()
                                        + File.separatorChar + config.getIndexPage());
                                    
                                    if (testFile.exists()) {
                                        file = testFile;
                                        contentType = findMimeType(config.getIndexPage());
                                    } else {
                                        if (!useDirListPage) {
                                            statusCode = FORBIDDEN_STATUS_CODE;
                                        }    
                                    }
                                } else {
                                    contentType = findMimeType(requestFile);
                                }
                            }
                        }

                        if (isLogging) {
                            String[] itemsToLog =
                                new String[] {
                                    socket.getInetAddress().getHostAddress(), line, statusCode
                                };
                            logger.log(itemsToLog);
                        }

                        if (statusCode.equals(BAD_STATUS_CODE)) {
                            return;
                        }
                        
                        headers.append(statusCode + " "
                            + String.valueOf(statusCodes.get(statusCode)));
                        headers.append(CRLF);
                        headers.append("Server: ");
                        headers.append(VERSION);
                        headers.append(CRLF);
                        headers.append("Content-Type: ");
                        headers.append(contentType);
                        headers.append(CRLF);
                        headers.append("Content-Length: ");

                        serverResponse = new DataOutputStream(socket.getOutputStream());
                        serverResponse.writeBytes(headers.toString());

                        if (statusCode.equals(OK_STATUS_CODE)) {
                            if (file.isDirectory()) {
                                String page = getDirListPage(file.list(), requestFile);
                                
                                serverResponse.writeBytes(getContentLengthResponse(page.length()));
                                serverResponse.writeBytes(page);
                            } else {
                                byte[] bytes = new byte[1024];
                                int bytesLen = bytes.length;
                                int offset;

                                fileInputStream =
                                    new BufferedInputStream(new FileInputStream(file));
                                
                                serverResponse.writeBytes(
                                    getContentLengthResponse(fileInputStream.available()));

                                while ((offset = fileInputStream.read(bytes, 0, bytesLen)) != -1) {
                                    serverResponse.write(bytes, 0, offset);
                                }
                            }
                        } else {
                            String page = getErrorPage(statusCode, unparsedRequestFile);
                            
                            serverResponse.writeBytes(getContentLengthResponse(page.length()));
                            serverResponse.writeBytes(page);
                        }
                    }
                }
            } catch (Exception ex) {
                ;
            } finally {
                if (fileInputStream != null) {
                    try {
                        fileInputStream.close();
                    } catch (IOException ex) {
                        ;
                    }
                }

                if (clientRequest != null) {
                    try {
                        clientRequest.close();
                    } catch (IOException ex) {
                        ;
                    }
                }

                if (serverResponse != null) {
                    try {
                        serverResponse.close();
                    } catch (IOException ex) {
                        ;
                    }
                }

                if (socket != null) {
                    try {
                        socket.close();
                    } catch (IOException ex) {
                        ;
                    }
                }
            }
        }
        
        private String getContentLengthResponse(int length) {
            return length + CRLF + CRLF;    
        }

        private String getDirListPage(String[] files, String path) {
            int startLen = START_LOOP_DIRECTIVE.length();
            int startPos = dirListTemplate.indexOf(START_LOOP_DIRECTIVE);
            int endLen = END_LOOP_DIRECTIVE.length();
            int endPos = dirListTemplate.indexOf(END_LOOP_DIRECTIVE);
            StringBuffer result = new StringBuffer(dirListTemplate);
            String listing = "";

            if (files.length > 0) {
                String newPath = ((path.endsWith("/")) ? path : path + '/');
                String innerElement = dirListTemplate.substring((startPos + startLen), endPos);
                StringBuffer lines = new StringBuffer();

                Arrays.sort(files);

                for (int i = 0; i < files.length; i++) {
                    lines.append(innerElement.replaceAll(ITEM_LINK_DIRECTIVE, newPath + files[i])
                        .replaceAll(ITEM_NAME_DIRECTIVE, files[i]));
                }

                listing = lines.toString();
            }

            result = result.replace(startPos, (endPos + endLen), listing);

            return result.toString().replaceAll(DIRECTORY_DIRECTIVE, path);
        }

        private String getErrorPage(String statusCode, String requestedUrl) {
            String statusCodeMessage = String.valueOf(statusCodes.get(statusCode));
            String result = statusCode + " " + statusCodeMessage;

            if (useErrorPage) {
                result = errorTemplate;
                result = result.replaceAll(STATUS_CODE_DIRECTIVE, statusCode);
                result = result.replaceAll(STATUS_CODE_MESSAGE_DIRECTIVE, statusCodeMessage);
                result = result.replaceAll(REQUESTED_URL_DIRECTIVE, requestedUrl);
            }

            return result;
        }

        private String findMimeType(String requestFile) {
            Map mimeTypes = config.getMimeTypes();
            String mimeType = DEFAULT_MIME_TYPE;
            int pos = requestFile.lastIndexOf('.');
            String ext = requestFile.substring(pos + 1, requestFile.length());

            for (Iterator it = mimeTypes.keySet().iterator(); it.hasNext();) {
                String key = (String) it.next();
                String value = (String) mimeTypes.get(key);

                if (value.indexOf(ext) > -1) {
                    mimeType = key;
                    break;
                }
            }

            return mimeType;
        }

        private String parseRequestFile(String requestFile) {
            try {
                requestFile = URLDecoder.decode(requestFile, "UTF-8");
            } catch (UnsupportedEncodingException ex) {
                ;
            }

            int pos = requestFile.lastIndexOf('?');

            if (pos > -1) {
                requestFile = requestFile.substring(0, pos);
            }

            return requestFile;
        }
    }
}
